# requires bash

cc_share="/usr/share/cubbli-conf"
cc_lib="$cc_share/lib"
cc_profiles="$cc_share/profiles"
cc_etc="/etc/cubbli/cubbli-conf"
cc_var="/var/lib/cubbli-conf"

cc_default="generic/generic"

cc_pf_bak="~cc-bak"
cc_pf_del="~cc-del"
cc_hash_held="disabled"	# legacy, should be "held" or something like that

# Compatibility levels.
#
#  0: no compatibility level set (reserved)
#  1: first deployment of cubbli-conf (not used)
#  2: cubbli-conf makes backups of overwritten files
#
cc_compat="2"


# default profile
cc_fn_default() {
	RET="$cc_etc/default"
	test -r "$RET" && return 0 || return 1
}


# deployed files
cc_fn_files() {
	local pkg="$1"

	RET="$cc_var/$pkg.files"
	test -f "$RET" && return 0 || return 1
}


# used profile for package
cc_fn_current_profile() {
	local pkg="$1"

	RET="$cc_var/$pkg.profile"
	test -f "$RET" && return 0 || return 1
}


# custom profile for package
cc_fn_want_profile() {
	local pkg="$1"

	RET="$cc_etc/$pkg.profile"
	test -f "$RET" && return 0 || return 1
}


# saved hash
cc_fn_saved_hash() {
	RET="$cc_var/sha256sums"
	test -f "$RET" && return 0 || true
	touch "$RET"
	return 1
}


# hashes to be updated
cc_fn_post_hash() {
	RET="$cc_var/sha256sums.post"
	test -f "$RET" && return 0 || return 1
}


# package prerc
cc_fn_prerc() {
	local pkg="$1"
	local prof="$2"

	RET="$cc_profiles/$pkg/$prof.prerc"
	test -f "$RET" && return 0 || return 1
}


# package postrc
cc_fn_postrc() {
	local pkg="$1"
	local prof="$2"

	RET="$cc_profiles/$pkg/$prof.postrc"
	test -f "$RET" && return 0 || return 1
}


# Fet compatibility version filename.
#
#	R: filename
#	r: 0 if found, 1 if not
#
cc_fn_compat() {
	RET="$cc_var/compat"
	test -f "$RET" && return 0 || return 1
}


# profile directory
cc_dn_profile() {
	local pkg="$1"
	local prof="$2"

	RET=$(echo "$cc_profiles/$pkg/$prof" | sed 's|///*|/|g')
	test -d "$RET" && return 0 || return 1
}


# package directory
cc_dn_package() {
	local pkg="$1"

	RET="$cc_profiles/$pkg"
	test -d "$RET" && return 0 || return 1
}


# Get compatibility level.
#
#	R: compat version, 0 if unset
#
cc_get_compat() {
	if ! cc_fn_compat; then
		RET=0
	else
		RET=$(cat $RET)
	fi
	return 0
}


# Set compatibility level.
#
#	1: version
#	r: 0 on success
#
cc_set_compat() {
	local level="$1"

	if ! test "$level"; then
		echo "cc_set_compat: Compat level is empty!" >&2
		exit 9
	fi

	test "$CC_VERBOSE" && echo "Setting compatibility level to $level"

	cc_fn_compat || true
	if ! test "$CC_SIMULATE"; then
		if ! echo "$level" >"$RET"; then
			echo "$RET: Cannot write compatibility level" >&2
			return 1
		fi
	else
		echo "simulate: Compatibility level set to $level"
	fi
}


# Check compatibility level.
#
#	r: 0 if up to date, 1 if not
#
cc_check_compat() {
	cc_get_compat
	test "$RET" = "$cc_compat" && return 0 || return 1
}


# Upgrade compatibility level.
#
cc_upgrade_compat() {
	cc_get_compat
	local n=$(( $RET + 1 ))

	# Not required as "seq" will do nothing if compat level is up to date.
	# It's just cleaner this way.
	cc_check_compat && return 0

	test ! "$CC_VERBOSE" || echo "Upgrading compatibility level..."

	local sim
	test "$CC_SIMULATE" && sim="echo simulate: " || sim=

	local l
	for l in $(seq $n $cc_compat); do
		test ! "$CC_VERBOSE" || echo "Upgrading to level $l..."
		if ! $sim sh $cc_lib/upgrade-compat-$l.sh; then
			echo "Upgrading to compatibility level $l failed!" >&2
			exit 9
		else
			cc_set_compat $l || exit 9
		fi
	done
}


# Remove file.
#
#	1: file
#	r: 0 on success
#
cc_rm() {
	local file="$1"

	if test "$CC_SIMULATE"; then
		echo "simulate: rm $file"
		return 0
	fi

	if ! rm "$file"; then
		echo "$file: Failed to remove" >&2
		ls -l "$file" 3>&2 2>/dev/null 1>&3 || true

		if ! test $CC_FORCE; then
			echo "Use -f to ignore." >&2
			exit 1
		fi
		return 1
	else
		return 0
	fi
}


# Replace file.
#
#	1: new
#	2: old
#	r: 0 on success
#
cc_mv_replace() {
	local new="$1"
	local old="$2"

	if test "$CC_SIMULATE"; then
		echo "simulate: mv $new $old"
		return 0
	fi

	if ! mv "$new" "$old"; then
		echo "$old: Failed to replace with $new" >&2
		ls -l "$new" 3>&2 2>/dev/null 1>&3 || true
		ls -l "$old" 3>&2 2>/dev/null 1>&3 || true

		if ! test $CC_FORCE; then
			echo "Use -f to ignore." >&2
			exit 1
		fi
		return 1
	else
		return 0
	fi
}


cc_package_exists() {
	local pkg="$1"

	test -d "$cc_profiles/$pkg" && return 0 || return 1
}


# get default profile
cc_get_default() {
	local file="$cc_etc/default"
	if ! test -r "$file"; then
		RET="generic/generic"
		return 1
	fi
	RET=$(cat "$file")
	return 0
}


# Print files in given package/profile.
#
#	1: package
#	2: profile
#	r: 0 if profile exists.
#
cc_get_profile_files() {
	local pkg="$1"
	local prof="$2"

	cc_dn_profile "$pkg" "$prof" || return 1
	local dir="$RET"

	(cd "$dir"; find -L . -type f | cut -c 2-)
	return 0
}


# set wanted profile
cc_set_profile() {
	local pkg="$1"
	local prof="$2"

	cc_fn_want_profile "$pkg" || true
	if ! echo "$prof" >"$RET"; then
		echo "$pkg: Failed to set package profile" >&2
		return 1
	fi

	return 0
}


# unset wanted profile
cc_unset_profile() {
	local pkg="$1"

	if cc_fn_want_profile "$pkg"; then
		if ! rm -f "$RET"; then
		       echo "$pkg: Failed to set profile to default" >&2
		       return 1
	       fi
	fi

	return 0
}


# get profile that is in use
cc_get_current_profile() {
	local pkg="$1"

	cc_fn_current_profile "$pkg" || return 1
	RET=$(cat "$RET")

	return 0
}


# get profile that user would want to use
cc_get_wanted_profile() {
	local pkg="$1"

	cc_fn_want_profile "$pkg" || return 1
	RET=$(cat "$RET")

	return 0
}


cc_profile_helper() {
	local pkg="$1"
	local prof="$2"

	# first try profile as-is
	if cc_dn_profile "$pkg" "$prof"; then
		RET="$prof"
		return 0
	fi

	# then replace minor name with generic (foo/bar -> foo/generic)
	prof=$(echo "$prof" | sed 's|/.*|/generic|')
	if cc_dn_profile "$pkg" "$prof"; then
		RET="$prof"
		return 0
	fi

	return 1
}


# get profile for a package
cc_get_profile() {
	local pkg="$1"
	local failsafe="$2"
	local prof

	# should we use per-package profile?
	if cc_get_wanted_profile "$pkg"; then
		cc_profile_helper "$pkg" "$RET" && return 0
	fi

	# try default profile
	if cc_get_default; then
		cc_profile_helper "$pkg" "$RET" && return 0
	fi

	# fine. falling back to generic/generic which MUST exist!
	prof="generic/generic"

	if ! cc_dn_profile "$pkg" "$prof"; then
		if test ! "$failsafe"; then
			echo "Generic profile missing" >&2
			echo "$pkg configuration package is BROKEN!" >&2
			exit 9
		fi
	fi
	RET="$prof"

	return 0
}


# Get list of hashed files.
#
#	r: 0 on success
#
cc_get_hashed_files() {
	cc_fn_saved_hash || return 0
	sed 's/^[0-9a-f]*  //' "$RET"
	return 0
}


# Get (calculate) hash for file.
#
#	1: file
#	R: hash
#
cc_get_hash() {
	local file="$1"

	if ! RET=$(sha256sum "$file"); then
		echo "$file: Can't get hash" >&2
		RET="Can't hash"
		return 1
	fi
	RET=$(echo "$RET" | cut -d ' ' -f 1)
	return 0
}


cc_get_hash_helper() {
	local list="$1"
	local file="$2"

	RET=$(grep " $file$" "$list" | cut -d ' ' -f 1)
	return 0
}


# Get saved hash.
#
#	1: file
#	R: hash
#
cc_get_saved_hash() {
	local file="$1"

	if ! cc_fn_saved_hash; then
		RET=""
		return 1
	else
		cc_get_hash_helper "$RET" "$file"
		return 0
	fi
}


# Remove hash for file.
#	1: file
#	r: 0 on success
#
cc_remove_hash() {
	local file="$1"

	cc_fn_saved_hash || return 0
	local hashfile="$RET"

	if ! sed -i "\| $file$| d" "$hashfile"; then
		echo "$hashfile: Failed to remove entry" >&2
		return 1
	else
		return 0
	fi
}


# Save hash in file.
#
#	1: file
#	2: hash (will be calculated if hash is empty)
#	r: 0 = ok, 1 = error
#
cc_save_hash() {
	local file="$1"
	local sum="$2"

	cc_fn_saved_hash || true
	local hashfile="$RET"

	cc_get_hash_helper "$hashfile" "$file" || true
	local oldhash="$RET"

	if ! test "$sum"; then
		cc_get_hash "$file" || return 1
		sum="$RET"
	fi

	test "$sum" = "$oldhash" && return 0

	cc_remove_hash "$file" || return 1

	if ! echo "$sum  $file" >>"$hashfile"; then
		echo "$hashfile: Failed to update entry" >&2
		return 1
	fi

	return 0
}


cc_hold_file() {
	local file="$1"
	cc_save_hash "$file" "$cc_hash_held"
}


cc_unhold_file() {
	local file="$1"
	cc_remove_hash "$file"
}


cc_run_posthash() {
	cc_fn_post_hash || return 0
	local list="$RET"

	local r=0
	local file
	while read file; do	# <"$list"
		if ! cc_save_hash "$file"; then
			echo "$file: Failed to update hash list" >&2
			r=1
		fi
	done <"$list"

	if ! rm -f "$list"; then
		echo "$list: Cannot remove post-hash list" >&2
		return 1
	fi

	return $r
}


# Check if file can be replaced. cc_can_replcae takes care of CC_FORCE.
#
#	1: file
#	r: 0 = yes, 1 = no
#
cc_can_replace() {
	local file="$1"
	local r=0

	# In short: File can be replaced only if it has not been changed
	#	locally or if action is forced.
	#
	# In detail (ignoring force):
	#	if (hash_for_file(file))
	#		if (hash_implies_keep(file))
	#			return false
	#		else if (exists(file))
	#			return files_match()
	#		else
	#			return false
	#	else
	#		return true

	cc_get_saved_hash "$file" || true
	local hash_saved="$RET"

	local what
	test "$CC_FORCE" && what="overriding" || what="skipping"

	if test "$hash_saved"; then
		if test "x$hash_saved" = "x$cc_hash_held"; then
			test "$CC_VERBOSE" && echo "$file: Held back - $what"
			test "$CC_FORCE" && return 0 || return 1
		elif test -e "$file"; then
			cc_get_hash "$file" || r=1
			local hash_now="$RET"
			if test "x$hash_saved" = "x$hash_now"; then
				return 0
			else
				test "$CC_VERBOSE" && \
					echo "$file: Locally edited - $what"
				test "$CC_FORCE" && return 0 || return 1
			fi
		else
			test "$CC_VERBOSE" && \
				echo "$file: Locally deleted - $what"
			test "$CC_FORCE" && return 0 || return 1
		fi
	else
		return 0
	fi
}


# Restore single file.
#
#	1: file
#	r: 0 if file was restored, 1 if file is held or locally modified,
#	   2 if restoration failed
#
cc_restore_file() {
	# simulation ok

	local file="$1"
	local ret=0

	# cc_can_replace takes care of CC_FORCE
	cc_can_replace "$file" || return 1

	if test -f "$file$cc_pf_del"; then
		cc_rm "$file" || ret=2
		cc_rm "$file$cc_pf_del" || ret=2
		if test "$CC_VERBOSE"; then
			if test $ret -eq 0; then
				echo "$file: Restored (deleted)"
			else
				echo "$file: Restore (delete) failed"
			fi
		fi
	elif test -f "$file$cc_pf_bak"; then
		cc_mv_replace "$file$cc_pf_bak" "$file" || ret=2
		if test "$CC_VERBOSE"; then
			if test $ret -eq 0; then
				echo "$file: Restored"
			else
				echo "$file: Restore failed"
			fi
		fi
	else
		echo "$file: No file to restore" >&2
		test $CC_FORCE || ret=2
	fi

	if test $ret -eq 0; then
		cc_remove_hash "$file" || return 2
	fi

	return $ret
}


# Restore original files.
#
#	1: package
#	2: new profile or empty, if restoring all files
#	r: 0 on success
#
cc_restore() {
	# simulation ok
	local pkg="$1"
	local prof="$2"
	local ret=0

	if ! cc_fn_files "$pkg"; then
		test "$CC_VERBOSE" && echo "$pkg: No files deployed" >&2
		return 0
	fi

	local old_files="$RET"
	local new_files=$(mktemp)
	local tobe_files=$(mktemp)

	if test "$prof"; then
		cc_get_profile_files "$pkg" "$prof" >"$tobe_files" || \
				true
	fi

	# Technically we could get a list of files to be restored
	# with "comm -23 old new".

	# Useless use of cat. I just think it makes it more readable.
	while read file; do	# <"$old_files"
		local insert=0
		if grep "^$file$" "$tobe_files" >/dev/null; then
			# New file set includes this file. Don't restore.
			insert=1
		else
			# We're not going to install this file. Restore it.
	
			local r=0
			# Ugly, but apparently the only way to get the return
			# code without "set -e" getting upset about non-zero.
			#
			# Non-zero return code means file was not restored.
			# RC 1 is not an error but a condition. Anything greater
			# than 1 is an error (and anything equal or lower than 1
			# is ok).
			#
			cc_restore_file "$file" || r=$?
			test $r -eq 0 || insert=1
			test $r -le 1 || ret=1
		fi
		if test $insert -eq 1; then
			if ! echo "$file" >>"$new_files"; then
				echo "$file: Cannot add to file list!" >&2
				ret=2
			fi
		fi
	done <"$old_files"

	if test -s "$new_files"; then
		if ! cc_mv_replace "$new_files" "$old_files"; then
			echo "$pkg: Failed to update packet file list" >&2
			echo "Old file list: $old_files" >&2
			echo "New file list: $new_files" >&2
			ret=2
		fi
	else
		cc_rm "$old_files" || ret=1
		cc_rm "$new_files" || ret=1
	fi

	cc_rm "$tobe_files" || ret=1

	# Clean up after simulation
	if test "$CC_SIMULATE"; then
		rm "$new_files"
		rm "$tobe_files"
	fi

	return $ret
}


# Make back up of current file.
#
#	1: file
#	r: 0 on success
#
cc_make_backup() {
	local file="$1"

	local sim
	test "$CC_SIMULATE" && sim="echo simulate: " || sim=

	if test -f "$file"; then
		local target="$file$cc_pf_bak"
		if test -e "$target" -a ! -f "$target"; then
			echo "$target: Already exists!" >&2
			return 1
		fi

		if ! $sim mv "$file" "$target"; then
			echo "$file: Failed to create back up" >&2
			return 1
		fi
	else
		local target="$file$cc_pf_del"

		if ! mkdir -p $(dirname "$target"); then
			echo "$target: Cannot create directory!" >&2
			return 1
		fi

		if test -e "$target" -a ! -f "$target"; then
			echo "$target: Already exists!" >&2
			return 1
		fi

		if ! $sim touch "$target"; then
			echo "$file: Failed to create back up" >&2
			return 1
		fi
	fi

	test "$CC_VERBOSE" && echo "$file: Backed up" || true

	return 0
}


# Check the type of overwritten file.
#
#	1: file
#	r: 0 = backup found, 1 = missing
#	R: status
#
cc_check_backup() {
	local file="$1"

	if test -f "$file$cc_pf_bak"; then
		RET="restore"
		return 0
	elif test -f "$file$cc_pf_del"; then
		RET="delete"
		return 0
	else
		RET="missing"
		return 1
	fi
}


cc_update() {
	local pkg="$1"

	cc_get_profile "$pkg"
	local prof="$RET"

	if ! cc_dn_profile "$pkg" "$prof"; then
		echo "$pkg: Cannot deploy, profile $prof missing!" >&2
		return 1
	fi
	local dir="$RET"

	# Restore old files.
	cc_restore "$pkg" "$prof" || echo "   Continuing despite errors." >&2

	cc_fn_files "$pkg" || true
	local files="$RET"

	cc_fn_saved_hash || true
	local hashfile="$RET"

	cc_fn_post_hash || true
	local posthash="$RET"


	# Technically we could make backups beforehand.
	# To get a list of files, "comm -13 old new".

	local r=0
	local file
	(cd "$dir"; find -L . -type f | cut -c 2-) |
	while read file; do
		local src="$dir/$file"

		# cc_can_replace takes care of CC_FORCE
		if ! cc_can_replace "$file"; then
			if ! grep "^$file$" "$files" >/dev/null 2>&1; then
				if ! echo "$file" >>"$files"; then
					echo "$file: Cannot add to file list" \
							>&2
					r=1
				fi
			fi
			continue
		fi

		if ! cc_check_backup "$file"; then
			cc_make_backup "$file" || r=1
		fi

		test "$CC_VERBOSE" && echo "$file: Deployed (using $src)"
		local tdir=$(dirname "$file")
		test -d "$tdir" || mkdir -p "$tdir"
		if cp -aL "$src" "$file"; then
			cc_save_hash "$file" || r=1
			grep "^$file$" "$files" >/dev/null 2>&1 ||
					echo "$file" >>"$files"
			echo "$file" >>"$posthash"
		else
			echo "$file: Failed to copy: $src" >&2
			r=1
		fi
	done

	cc_fn_current_profile "$pkg" || true
	if ! echo "$prof" >"$RET"; then
		echo "$pkg: Failed to update profile record: $RET" >&2
		r=1
	fi

	return $r
}


cc_prerc() {
	local pkg="$1"

	cc_get_profile "$pkg"
	local prof="$RET"

	cc_fn_prerc "$pkg" "$prof" || return 0
	local cmd="$RET"
	RET=""
	"$cmd" && return 0 || return 1
}


cc_postrc() {
	local pkg="$1"

	cc_get_profile "$pkg"
	local prof="$RET"

	cc_fn_postrc "$pkg" "$prof" || return 0
	local cmd="$RET"
	RET=""
	"$cmd" && return 0 || return 1
}
